static void *rcs_id = (rcs_id, "$Header: /LocalDeveloper/Source/CVS/OmniGroup/bb/server/changeroom.c,v 1.4 1993/08/13 05:08:31 kc Exp $");

#include "server.h"

void server_changeroom(int i)
{
    char room[ROOM_SIZE];
    char owner[ID_SIZE];
    bb_uid owner_uid;
    time_t default_expiration;
    class_t default_class;
    char summary_line[LINE_SIZE];
    directory_entry room_info;
    bb_rid rid;
    bb_priority priority;
    bb_archive_t archive;

    read_socket_s(Client[i].socket, room, sizeof(room));
    read_socket_s(Client[i].socket, owner, sizeof(owner));
    read_socket_int(Client[i].socket, (int *)&default_expiration);
    read_socket_int(Client[i].socket, (int *)&default_class);
    read_socket_s(Client[i].socket, summary_line, sizeof(summary_line));
    read_socket_int(Client[i].socket, (int *)&priority);
    if (Client[i].minor_version >= 1) {
	read_socket_int(Client[i].socket, (int *)&archive);
    }
    if ((rid = get_rid(room)) == -1)
	SERVER_ERROR("no such room");
    if (!get_directory_entry(&room_info, rid))
	SERVER_ERROR("error accessing room information");
    if (Client[i].uid != room_info.owner_uid && !Client[i].owner)
	SERVER_ERROR("permission denied");
    if ((owner_uid = get_uid(owner)) == -1)
	SERVER_ERROR("room information not changed: no such user");
    if (room_info.owner_uid != owner_uid) {
	char subject[80];
	FILE *mail_message;

	sprintf(subject, "You now own the %s room on bb", room_info.name);
	mail_message = mail(Client[i].uid, owner_uid, subject);
	if (!mail_message)
	    SERVER_ERROR("error notifying new room owner");
	fprintf(mail_message, "%s transferred ownership of the %s room to %s.",
		Client[i].id, room_info.name, get_id(owner_uid));
	pclose(mail_message);
	room_info.owner_uid = owner_uid;
    }
    if (default_expiration != -1) {
	if (default_expiration > expiration_limit(room_info.groups) &&
	    !Client[i].owner)
	    SERVER_ERROR("room information not changed: expiration too long");
	if (default_expiration < MIN_EXPIRATION && !Client[i].owner)
	    SERVER_ERROR("room information not changed: expiration too short");
	room_info.default_expiration = default_expiration;
    }
    if (default_class != -1)
	room_info.default_class = default_class;
    bcopy(summary_line, room_info.summary_line,
	  sizeof(room_info.summary_line));
    if (priority) {
	if (priority > 0)
	    priority--;
	if (priority > 0 && !Client[i].owner)
	    SERVER_ERROR("the offset priority must be negative");
	room_info.room_offset = MAX(priority, SYSTEM_CUTOFF);
    }
    if (Client[i].minor_version >= 1) {
	if (archive != -1) {
	    if (archive == no_archive || room_info.archive == no_archive) {
		char *archives[] =
		{"common", "mail", "none"};

		syslog(LOG_INFO,
		       "%s (%d) changed the archive of the %s (%d) room from %s to %s",
		       Client[i].id, Client[i].uid, room, rid,
		       archives[room_info.archive], archives[archive]);
	    }
	    room_info.archive = (bb_archive_t)archive;
	}
    }
    if (!put_directory_entry(rid, &room_info))
	SERVER_ERROR("error storing room information");
    write_socket_code(Client[i].socket, SERVER_ACK);
}
