static void *rcs_id = (rcs_id, "$Header: /LocalDeveloper/Source/CVS/OmniGroup/bb/server/database.c,v 1.6 1994/03/18 00:35:37 bb Exp $");

#include <fcntl.h>
#include <ndbm.h>
#include <sys/file.h>
#include "server.h"

#ifndef COPY
#ifdef X
#define COPY(src, dest, dtype) *dest = *((dtype *) src);
#else
#define COPY(src, dest, dtype) bcopy(src, dest, sizeof(dtype));
#endif
#endif

static FILE *dir_f, *user_f, *summary_f;
static DBM *room_dbm;
static summary_list *summary_cache;
static int max_messages;

int Directory_stamp, Group_directory_stamp;
bb_rid Last_rid = -1;
bb_gid Last_gid = -1;
bb_uid Last_uid = -1;
boolean Read_only;		/* whether messagebase is read-only */
int Max_active_slot = -1;	/* the maximum slot number currently in use */
struct client_struct Client[MAX_CLIENTS];

 /* an array containing information on each client connection */
time_t Now;
bb_uid Owner_uid = OWNER_UID, Daemon_uid = DAEMON_UID;

#define DATA_ROOM_ID 'i'
#define DATA_UID 'u'
#define DATA_RID 'r'
#define DATA_GID 'g'
#define DATA_GROUP 'G'

typedef struct {
    char type;
    bb_rid rid;
    bb_uid uid;
}      room_id;
typedef struct {
    char type;
    bb_rid rid;
    long message;
}      room_message;
typedef struct {
    char type;
    bb_gid gid;
}      key_group;
typedef struct {
    char type;
    char id[ID_SIZE];
}      data_key_id;
typedef struct {
    char type;
    char group[GROUP_SIZE];
}      data_key_gid;
typedef struct {
    char type;
    char name[ROOM_SIZE];
}      data_key_rid;

static char *rid_flag = "< last rid >", *gid_flag = "< last gid >", *uid_flag = "< last uid >";

int database_init(boolean read_only, boolean override_locks)
{
    Directory_stamp = Group_directory_stamp = 1;
    Read_only = read_only;
    if (!Read_only) {
	if (!(dir_f = fopen(DIRECTORY, "r+"))) {
	    dir_f = fopen(DIRECTORY, "w+");
	}
	if (dir_f && flock(fileno(dir_f), LOCK_EX | LOCK_NB) == -1 &&
	    !override_locks) {
	    fclose(dir_f);
	    return 0;
	}
	if (!(user_f = fopen(USER, "r+"))) {
	    user_f = fopen(USER, "w+");
	}
	if (!(summary_f = fopen(SUMMARY, "r+"))) {
	    summary_f = fopen(SUMMARY, "w+");
	}
	room_dbm = dbm_open(ROOM, O_RDWR | O_CREAT, 0666);
    } else {
	dir_f = fopen(DIRECTORY, "r");
	user_f = fopen(USER, "r");
	summary_f = fopen(SUMMARY, "r");
	room_dbm = dbm_open(ROOM, O_RDONLY, 0666);
    }
    if (room_dbm) {
	Last_rid = get_rid(rid_flag);
	Last_gid = get_gid(gid_flag);
	Last_uid = get_uid(uid_flag);
    }
    max_messages = 2048;	/*** this won't last forever ***/
    init_cache();
    if (!Read_only && Last_gid < 0) {
    /* This is a new database--perform some initialization */

	group_entry group_info;
	bb_gid gid;

	strcpy(group_info.name, "default");
	group_info.owner = Owner_uid;
	group_info.default_class = DEFAULT_GROUP_CLASS;
	group_info.group_offset = DEFAULT_GROUP_OFFSET;
	group_info.max_expiration = MAX_EXPIRATION;
	gid = add_gid(group_info.name);
	if (!put_group_entry(gid, &group_info)) {
	    syslog(LOG_ERR, "database_init: group initialization failed\n");
	    database_shutdown();
	    return FALSE;
	}
    }
    return (dir_f && user_f && summary_f && room_dbm);
}

void database_shutdown()
{
    dump_cache();
    flock(fileno(dir_f), LOCK_UN);
    fclose(dir_f);
    fclose(user_f);
    fclose(summary_f);
    dbm_close(room_dbm);
    if (summary_cache) {
	free(summary_cache);
	summary_cache = NULL;
    }
}

int get_user_entry(user_entry *user_info, bb_uid uid)
{
    if  (uid < 0)
	    return FALSE;
    fseek(user_f, uid * sizeof(user_entry), 0);
    if (fread(user_info, sizeof(user_entry), 1, user_f) != 1)
	return FALSE;
    if (user_info->uses == 0)
	return FALSE;
    return TRUE;
}

int put_user_entry(bb_uid uid, const user_entry *user_info)
{
    if  (uid < 0)
	    return FALSE;
    fseek(user_f, uid * sizeof(user_entry), 0);
    if (fwrite(user_info, sizeof(user_entry), 1, user_f) != 1)
	return FALSE;
    return TRUE;
}

int remove_user_entry(bb_uid uid)
{
    user_entry user_info;

    bzero(&user_info, sizeof(user_info));
    return put_user_entry(uid, &user_info);
}

int get_directory_entry(directory_entry *directory_info, bb_rid rid)
{
    if  (rid < 0)
	    return FALSE;
    fseek(dir_f, rid * sizeof(directory_entry), 0);
    return fread(directory_info, sizeof(directory_entry), 1, dir_f)== 1 &&
      directory_info->next_message_number > 0;
}

int put_directory_entry(bb_rid rid, const directory_entry *directory_info)
{
    if  (rid < 0)
	    return FALSE;
    fseek(dir_f, rid * sizeof(directory_entry), 0);
    if (fwrite(directory_info, sizeof(directory_entry), 1, dir_f)!=1)
	return FALSE;
    Directory_stamp++;
    update_cached_room_user_info(rid, -1);
    return TRUE;
}

int remove_directory_entry(bb_rid rid)
{
    directory_entry directory_info;

    bzero(&directory_info, sizeof(directory_info));
    return put_directory_entry(rid, &directory_info);
}

int get_room_user_entry(room_user_entry *room_user_info,
			    bb_rid rid, bb_uid uid)
{
    datum key, content;
    room_id room_id;

    bzero(&room_id, sizeof(room_id));
    room_id.type = DATA_ROOM_ID;
    room_id.rid = rid;
    room_id.uid = uid;
    key.dptr = (char *)&room_id;
    key.dsize = sizeof(room_id);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return FALSE;
    COPY(content.dptr, room_user_info, room_user_entry);
    return TRUE;
}

int put_room_user_entry(bb_rid rid, bb_uid uid,
			const room_user_entry *room_user_info)
{
    datum key, content;
    room_id room_id;
    int result;

    bzero(&room_id, sizeof(room_id));
    room_id.type = DATA_ROOM_ID;
    room_id.rid = rid;
    room_id.uid = uid;
    key.dptr = (char *)&room_id;
    key.dsize = sizeof(room_id);
    content.dptr = (char *)room_user_info;
    content.dsize = sizeof(room_user_entry);
    result = dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
    if (result)
	update_cached_room_user_info(rid, uid);
    return result;
}

int remove_room_user_entry(bb_rid rid, bb_uid uid)
{
    datum key;
    room_id room_id;

    bzero(&room_id, sizeof(room_id));
    room_id.type = DATA_ROOM_ID;
    room_id.rid = rid;
    room_id.uid = uid;
    key.dptr = (char *)&room_id;
    key.dsize = sizeof(room_id);
    return dbm_delete(room_dbm, key) >= 0;
}

int get_summary_index(bb_rid rid, int pnum)
{
    static int i = -1;
    summary_entry *entry;
    int last_message = summary_cache[rid].count;

    if (!last_message)
	return -1;
    entry = summary_cache[rid].list;
    if (i < 0 || i >= last_message) {
	for (i = 0; i < last_message && entry[i].number < pnum; i++);
    } else if (entry[i].number == pnum) {
	return i;
    } else if (entry[i].number < pnum) {
	for (; i <= last_message && entry[i].number < pnum; i++);
    } else {
	for (; i >= 0 && entry[i].number > pnum; i--);
    }
    return (i >= 0 && i < last_message && entry[i].number == pnum)
      ? i : -1;
}

int db_get_message_entry(message_entry *message_ptr, bb_rid rid, int message)
{
    int i;
    summary_list *summary;

    if (!(summary = get_summary(rid)))
	return FALSE;
    if ((i = get_summary_index(rid, message)) < 0)
	return FALSE;
    *message_ptr = summary->list[i].header;
    return TRUE;
}

int put_message_entry(bb_rid rid, int message, const message_entry *message_info)
{
    int i;
    summary_list *summary;

    if (!(summary = get_summary(rid)))
	return FALSE;
    if ((i = get_summary_index(rid, message)) < 0) {
	int j;

    /* Find the message just prior to where we'll be inserting this one. */

	for (i = summary->count - 1;
	     i >= 0 && summary->list[i].number > message;
	     i--);

    /* Move past it. */

	i++;

    /* Allocate the extra memory for this message. */

	summary->count++;
	summary->list = Realloc(summary->list, summary_entry, summary->count);

    /* Shift messages from i through the last message. */

	for (j = summary->count - 1; j > i; j--)
	    summary->list[j] = summary->list[j - 1];

    /* Put the message here! */

	summary->list[i].number = message;
    }
    summary->list[i].header = *message_info;
    summary->version++;
    return TRUE;
}

int remove_message_entry(bb_rid rid, int message)
{
    int i;
    summary_list *summary;

    if (!(summary = get_summary(rid)))
	return FALSE;
    if ((i = get_summary_index(rid, message)) < 0)
	return FALSE;
    summary->version++;
    summary->count--;
    for (; i < summary->count; i++)
	summary->list[i] = summary->list[i + 1];
    summary->list = Realloc(summary->list, summary_entry, summary->count);
    return TRUE;
}

const char *get_id(bb_uid uid)
{
    static user_entry user_info;

    return get_user_entry(&user_info, uid) ? user_info.id : "******";
}

bb_uid add_uid(const char *id)
{
    Last_uid++;
    put_uid(Last_uid, uid_flag);
    put_uid(Last_uid, id);
    return Last_uid;
}

int put_uid(bb_uid u, const char *id)
{
    datum key, content;
    data_key_id key_id;

    bzero(&key_id, sizeof(key_id));
    key_id.type = DATA_UID;
    downcase(key_id.id, id);
    key.dptr = (char *)&key_id;
    key.dsize = sizeof(key_id);
    content.dptr = (char *)&u;
    content.dsize = sizeof(u);
    return dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
}

bb_uid get_uid(const char *id)
{
    datum key, content;
    data_key_id key_id;
    bb_uid uid;

    bzero(&key_id, sizeof(key_id));
    key_id.type = DATA_UID;
    downcase(key_id.id, id);
    key.dptr = (char *)&key_id;
    key.dsize = sizeof(key_id);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return -1;
    COPY(content.dptr, &uid, bb_uid);
    return uid;
}

int remove_uid(const char *id)
{
    datum key;
    data_key_id key_id;

    bzero(&key_id, sizeof(key_id));
    key_id.type = DATA_UID;
    downcase(key_id.id, id);
    key.dptr = (char *)&key_id;
    key.dsize = sizeof(key_id);
    return dbm_delete(room_dbm, key) >= 0;
}

const char *get_roomname(bb_rid rid)
{
    static directory_entry room_info;

    return get_directory_entry(&room_info, rid) ? room_info.name : "";
}

bb_rid add_rid(const char *name)
{
    int i;

    Last_rid++;
    put_rid(Last_rid, rid_flag);
    put_rid(Last_rid, name);
    summary_cache = Realloc(summary_cache, summary_list, Last_rid + 1);
    bzero((char *)&summary_cache[Last_rid], sizeof(summary_list));
    for (i = 0; i <= Max_active_slot; i++)
	if (Client[i].dir_info) {
	    Client[i].dir_info = Realloc(Client[i].dir_info, directory_info *,
					 Last_rid + 1);
	    Client[i].dir_info[Last_rid] = NULL;
	}
    return Last_rid;
}

int put_rid(bb_rid r, const char *name)
{
    datum key, content;
    data_key_rid key_rid;

    bzero(&key_rid, sizeof(key_rid));
    key_rid.type = DATA_RID;
    downcase(key_rid.name, name);
    key.dptr = (char *)&key_rid;
    key.dsize = sizeof(key_rid);
    content.dptr = (char *)&r;
    content.dsize = sizeof(r);
    return dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
}

bb_rid get_rid(const char *name)
{
    datum key, content;
    data_key_rid key_rid;
    bb_rid rid;

    bzero(&key_rid, sizeof(key_rid));
    key_rid.type = DATA_RID;
    downcase(key_rid.name, name);
    key.dptr = (char *)&key_rid;
    key.dsize = sizeof(key_rid);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return -1;
    COPY(content.dptr, &rid, bb_rid);
    return rid;
}

int remove_rid(const char *name)
{
    datum key;
    data_key_rid key_rid;

    bzero(&key_rid, sizeof(key_rid));
    key_rid.type = DATA_RID;
    downcase(key_rid.name, name);
    key.dptr = (char *)&key_rid;
    key.dsize = sizeof(key_rid);
    return dbm_delete(room_dbm, key) >= 0;
}

const char *get_groupname(bb_gid gid)
{
    static group_entry group_info;

    return get_group_entry(&group_info, gid) ? group_info.name : "";
}

bb_gid get_gid(const char *group)
{
    datum key, content;
    data_key_gid key_gid;
    bb_gid g;

    bzero(&key_gid, sizeof(key_gid));
    key_gid.type = DATA_GID;
    downcase(key_gid.group, group);
    key.dptr = (char *)&key_gid;
    key.dsize = sizeof(key_gid);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return -1;
    COPY(content.dptr, &g, bb_gid);
    return g;
}

bb_gid add_gid(const char *name)
{
    Last_gid++;
    put_gid(Last_gid, gid_flag);
    put_gid(Last_gid, name);
    return Last_gid;
}

int put_gid(bb_gid gid, const char *group)
{
    datum key, content;
    data_key_gid key_gid;

    bzero(&key_gid, sizeof(key_gid));
    key_gid.type = DATA_GID;
    downcase(key_gid.group, group);
    key.dptr = (char *)&key_gid;
    key.dsize = sizeof(key_gid);
    content.dptr = (char *)&gid;
    content.dsize = sizeof(gid);
    return dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
}

int remove_gid(const char *group)
{
    datum key;
    data_key_gid key_gid;

    bzero(&key_gid, sizeof(key_gid));
    key_gid.type = DATA_GID;
    downcase(key_gid.group, group);
    key.dptr = (char *)&key_gid;
    key.dsize = sizeof(key_gid);
    return dbm_delete(room_dbm, key) >= 0;
}

int put_group_entry(bb_gid gid, const group_entry *group_info)
{
    datum key, content;
    key_group key_group;
    int result;

    Group_directory_stamp++;
    bzero(&key_group, sizeof(key_group));
    key_group.type = DATA_GROUP;
    key_group.gid = gid;
    key.dptr = (char *)&key_group;
    key.dsize = sizeof(key_group);
    content.dptr = (char *)group_info;
    content.dsize = sizeof(group_entry);
    result = dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
    if (result)
	update_cached_group_info(gid);
    return result;
}

int get_group_entry(group_entry *group_info, bb_gid gid)
{
    datum key, content;
    key_group key_group;

    bzero(&key_group, sizeof(key_group));
    key_group.type = DATA_GROUP;
    key_group.gid = gid;
    key.dptr = (char *)&key_group;
    key.dsize = sizeof(key_group);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return FALSE;
    COPY(content.dptr, group_info, group_entry);
    return TRUE;
}

int put_summary(bb_rid rid)
{
    int count = 0;

    if (Read_only)
	return FALSE;
    count = summary_cache[rid].count;
    if (count > max_messages) {
	syslog(LOG_ERR, "put_summary(%d): count overflow: %d\n", rid, count);
	return FALSE;
    }
    fseek(summary_f,
	  rid * (sizeof(int) + max_messages * sizeof(summary_entry)),
	  0);
    if (fwrite(&count, sizeof(count), 1, summary_f) != 1) {
	syslog(LOG_ERR, "put_summary(%d): writing count: %m\n", rid);
	return FALSE;
    }
    if (fwrite(summary_cache[rid].list, sizeof(summary_entry), count, summary_f)
	!= count) {
	syslog(LOG_ERR, "put_summary(%d): writing list: %m\n", rid);
	return FALSE;
    }
    summary_cache[rid].saved_version = summary_cache[rid].version;
    return TRUE;
}

summary_list *get_summary(bb_rid rid)
{
    summary_entry *summary;
    int count;

    if (rid > Last_rid) {
	syslog(LOG_ERR, "get_summary(%d): rid > Last_rid (%d)",
	       rid, Last_rid);
	return NULL;
    }
    if (!summary_cache)
	summary_cache[rid].last_access = Now;
    if (summary_cache[rid].version)
	return &summary_cache[rid];

    fseek(summary_f,
	  rid * (sizeof(int) + max_messages * sizeof(summary_entry)),
	  0);
    if (fread(&count, sizeof(count), 1, summary_f) != 1)
	count = 0;
    if (count > max_messages) {
	syslog(LOG_ERR, "get_summary(%d): invalid count: %d\n", count);
	return NULL;
    }
    summary_cache[rid].saved_version = 1;
    if (count > 0) {
	summary = (summary_entry *) malloc(count * sizeof(summary_entry));
	if (!summary) {
	    syslog(LOG_ERR, "get_summary(%d): malloc: %m\n", rid);
	    return NULL;
	}
	count = fread(summary, sizeof(summary_entry), count, summary_f);
	if (count == -1) {
	    syslog(LOG_ERR, "get_summary(%d): fread: %m\n", rid);
	    free(summary);
	    return NULL;
	}
    } else {
	summary = NULL;
    }
    summary_cache[rid].count = count;
    summary_cache[rid].list = summary;
    summary_cache[rid].version = 1;
    return &summary_cache[rid];
}

void init_cache()
{
    if   (summary_cache) {
	     free(summary_cache);
	summary_cache = NULL;
    }
    if (Last_rid >= 0) {
	summary_cache = (summary_list *)
	  malloc((Last_rid + 1) * sizeof(summary_list));
	bzero((char *)summary_cache, (Last_rid + 1) * sizeof(summary_list));
    } else {
	summary_cache = NULL;
    }
}

void flush_cache()
{
    bb_rid rid;

    for (rid = 0; rid <= Last_rid; rid++)
	if (summary_cache[rid].version &&
	    summary_cache[rid].saved_version < summary_cache[rid].version)
	    put_summary(rid);
}

void dump_cache()
{
    bb_rid rid;

    if (!summary_cache)
	return;
    for (rid = 0; rid <= Last_rid; rid++)
	if (summary_cache[rid].version) {
	    if (summary_cache[rid].saved_version < summary_cache[rid].version)
		put_summary(rid);
	    if (summary_cache[rid].list) {
		free(summary_cache[rid].list);
		summary_cache[rid].list = NULL;
	    }
	    summary_cache[rid].version = 0;
	}
}

void check_cache()
{
    bb_rid rid;

    for (rid = 0; rid <= Last_rid; rid++)
	if (summary_cache[rid].version) {
	    if (summary_cache[rid].saved_version < summary_cache[rid].version)
		put_summary(rid);
	    if (summary_cache[rid].list &&
		(Now - summary_cache[rid].last_access > CACHE_TIMEOUT ||
		 Max_active_slot < 0)) {
		free(summary_cache[rid].list);
		summary_cache[rid].list = NULL;
		summary_cache[rid].version = 0;
	    }
	}
}

void update_cached_room_user_info(bb_rid rid, bb_uid uid)
{
    int slot;

    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on &&
	    (Client[slot].uid == uid || uid == -1)) {
	    if (Client[slot].dir_info && Client[slot].dir_info[rid]) {
		free(Client[slot].dir_info[rid]);
		Client[slot].dir_info[rid] = NULL;
	    }
	    Client[slot].update_directory = 0;
	    if (Client[slot].stamp)
		Client[slot].stamp[rid].update++;
	}
}

boolean update_cached_group_info(bb_gid gid)
{
    int slot;
    group_entry group_info;

    if (!get_group_entry(&group_info, gid)) {
	ERROR("error accessing group information");
	return FALSE;
    }
    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on) {
	    user_entry userinfo;
	    bb_rid rid;

	    if (!get_user_entry(&userinfo, Client[slot].uid))
		continue;

	    Client[slot].group_priority[gid] =
	      userinfo.group_priority[gid] + group_info.group_offset;
	    Client[slot].group_class[gid] = userinfo.group_class[gid];

	    for (rid = 0; rid <= Last_rid; rid++) {
		if (Client[slot].dir_info[rid] &&
		    GROUP_ISSET(gid, &Client[slot].dir_info[rid]->groups)) {
		    free(Client[slot].dir_info[rid]);
		    Client[slot].dir_info[rid] = NULL;
		    Client[slot].stamp[rid].update++;
		}
	    }
	    Client[slot].update_directory = 0;
	}
    return TRUE;
}

void database_changelast(bb_gid lgid, bb_rid lrid, bb_uid luid)
{
         Last_gid = lgid;
    Last_rid = lrid;
    Last_uid = luid;
    put_gid(Last_gid, gid_flag);
    put_rid(Last_rid, rid_flag);
    put_uid(Last_uid, uid_flag);
    init_cache();
}
