#ifndef _BB_SERVER_DATABASE_H
#define _BB_SERVER_DATABASE_H

#include "../lib/types.h"

typedef long bb_uid;
typedef long bb_rid;
typedef long bb_gid;

typedef struct {
    char name[ROOM_SIZE];
    bb_uid owner_uid;
    time_t creation_time;
    time_t default_expiration;
    class_t default_class;
    long next_message_number;
    char summary_line[LINE_SIZE];
    bb_archive_t archive;
    long first_message_number;
    group_set groups;
    bb_priority room_offset;
} directory_entry;

typedef struct {
    char name[GROUP_SIZE];
    bb_uid owner;
    class_t default_class;
    bb_priority group_offset;
    time_t max_expiration;
} group_entry;

typedef struct {
    long last_message_read;
    bb_priority user_offset;
    class_t class;
} room_user_entry;

typedef struct {
    time_t date;
    time_t expiration;
    int size;
    int lines;
    boolean deleted;
    bb_uid poster_uid;
    char poster_nick[LINE_SIZE];
    char poster_id[ID_SIZE];
    char subject[LINE_SIZE];
} message_entry;

typedef struct {
    message_entry header;
    int number;
} summary_entry;

typedef struct {
    summary_entry *list;
    int count, version, saved_version;
    time_t last_access;
} summary_list;

typedef struct {
    time_t last_access_time;
    char id[ID_SIZE];
    char passwd[32];
    char nick[LINE_SIZE];
    char last_access_source[LINE_SIZE];
    char mail_address[LINE_SIZE];
    int uses;
    int posts;
    time_t connect_time;
    long UNUSED;
    bb_priority cutoff_priority;
    bb_priority group_priority[MAX_GROUPS];
    class_t group_class[MAX_GROUPS];
} user_entry;

/* Database variables, found in database.c */

extern bb_rid Last_rid;
extern bb_gid Last_gid;
extern bb_uid Last_uid;

extern boolean Read_only;
extern int Directory_stamp, Group_directory_stamp;
extern bb_uid Owner_uid, Daemon_uid;
extern time_t Now;

/* Database routines, found in database.c */

int database_init(boolean read_only, boolean override_locks);
void database_shutdown();

int get_user_entry(user_entry *user_info, bb_uid uid);
int put_user_entry(bb_uid uid, const user_entry *user_info);
int remove_user_entry(bb_uid uid);

int get_directory_entry(directory_entry *directory_info, bb_rid rid);
int put_directory_entry(bb_rid rid, const directory_entry *directory_info);
int remove_directory_entry(bb_rid rid);

int get_room_user_entry(room_user_entry *room_user_info,
			bb_rid rid, bb_uid uid);
int put_room_user_entry(bb_rid rid, bb_uid uid,
			const room_user_entry *room_user_info);
int remove_room_user_entry(bb_rid rid, bb_uid uid);

int db_get_message_entry(message_entry *message_ptr, bb_rid rid, int message);
int put_message_entry(bb_rid rid, int message, const message_entry *message_info);
int remove_message_entry(bb_rid rid, int message);

int get_group_entry(group_entry *group_info, bb_gid gid);
int put_group_entry(bb_gid gid, const group_entry *group_info);
int remove_group_entry(bb_gid gid);

const char *get_id(bb_uid uid);
bb_uid get_uid(const char *id);
bb_uid add_uid(const char *id);
int put_uid(bb_uid uid, const char *id);
int remove_uid(const char *id);

const char *get_roomname(bb_rid rid);
bb_rid get_rid(const char *name);
bb_rid add_rid(const char *name);
int put_rid(bb_rid rid, const char *name);
int remove_rid(const char *name);

const char *get_groupname(bb_gid gid);
bb_gid get_gid(const char *name);
bb_gid add_gid(const char *name);
int put_gid(bb_gid gid, const char *name);
int remove_gid(const char *name);

summary_list *get_summary(bb_rid rid);
int put_summary(bb_rid rid);

void init_cache();
void flush_cache();
void dump_cache();
void check_cache();

/* this is not for general use! */
void database_changelast(bb_gid lgid, bb_rid lrid, bb_uid luid);

#endif _BB_SERVER_DATABASE_H
