static void *rcs_id = (rcs_id, "$Header: /LocalDeveloper/Source/CVS/OmniGroup/bb/server/lib.c,v 1.4 1993/08/13 05:08:33 kc Exp $");

#include "server.h"
#include <errno.h>

char error_text[LINE_SIZE];
boolean Disable_expire = FALSE;

static boolean get_directory_info(directory_info *dir_info, int i, bb_rid rid);
static int archive(bb_rid rid, int message);

void update_dir_info(int i, bb_rid rid)
{
    if(!Client[i].dir_info[rid])
	Client[i].dir_info[rid] = (directory_info *)
	  malloc(sizeof(directory_info));
    if (!get_directory_info(Client[i].dir_info[rid], i, rid))
	*Client[i].dir_info[rid]->name = '\0';
}

void update_room_pointer(int i, bb_rid rid, int last_message_read)
{
    if(rid == Client[i].rid)
	Client[i].last_message_read = last_message_read;
    Client[i].update_directory = 0;
    Client[i].stamp[rid].update++;
    if (Client[i].dir_info[rid]) {
	summary_list *summary;
	summary_entry *entry;
	int new;

	if (!(summary = get_summary(rid)))
	    return;
	if (summary->count > 0) {
	    for (new = 0, entry = summary->list + summary->count - 1;
		 entry >= summary->list && entry->number > last_message_read;
		 entry--)
		if (!entry->header.deleted)
		    new++;
	} else
	    new = 0;
	Client[i].dir_info[rid]->new_messages = new;
    } else {
	update_dir_info(i, rid);
    }
}

boolean save_room_pointer(int i)
{
    room_user_entry room_user_info;
    bb_rid rid = Client[i].rid;

    if (rid == -1)
	return TRUE;
    if (!get_room_user_entry(&room_user_info, rid, Client[i].uid)) {
	room_user_info.user_offset = DEFAULT_USER_OFFSET;
	room_user_info.class = no_class;
    }
    room_user_info.last_message_read = Client[i].last_message_read;
    if (!put_room_user_entry(rid, Client[i].uid, &room_user_info)) {
	ERROR("error storing room pointers");
	return FALSE;
    }
    return TRUE;
}

directory_info *get_dir_info(int i, bb_rid rid)
{
    if(!Client[i].dir_info) {
	static directory_info dir_info;

	return get_directory_info(&dir_info, i, rid) ? &dir_info : NULL;
    }
    if (!Client[i].dir_info[rid])
	update_dir_info(i, rid);
    return *Client[i].dir_info[rid]->name ? Client[i].dir_info[rid] : NULL;
}

static boolean get_directory_info(directory_info *dir_info, int i, bb_rid rid)
{
    directory_entry room_info;
    room_user_entry room_user_info;
    summary_list *summary;
    summary_entry *entry;
    int last_message_read, new = 0;

    if (!get_directory_entry(&room_info, rid)) {
	ERROR("error accessing room information");
	return FALSE;
    }
    bcopy(room_info.name, dir_info->name, sizeof(dir_info->name));
    bcopy(room_info.summary_line, dir_info->summary,
	  sizeof(dir_info->summary));
    strcpy(dir_info->owner, get_id(room_info.owner_uid));
    if (!(summary = get_summary(rid))) {
	ERROR("reading summary");
	return FALSE;
    }
    dir_info->messagenumber = summary->count > 0 ?
      summary->list[summary->count - 1].number : 0;
    dir_info->default_expiration = room_info.default_expiration;
    dir_info->default_class = room_info.default_class;
    if (!get_room_user_entry(&room_user_info, rid, Client[i].uid)) {
	last_message_read = 0;
	dir_info->priority = DEFAULT_USER_OFFSET;
	dir_info->class = no_class;
    } else {
	int j;

	last_message_read = rid == Client[i].rid ?
	  Client[i].last_message_read : room_user_info.last_message_read;
	for (j = 0;
	     j < summary->count &&
	     summary->list[j].number <= room_user_info.last_message_read;
	     j++);
	dir_info->new_messages = summary->count - j;
	dir_info->priority = room_user_info.user_offset;
	dir_info->class = room_user_info.class;
    }
    if (summary->count > 0) {
	for (entry = summary->list + summary->count - 1;
	     entry >= summary->list && entry->number > last_message_read;
	     entry--)
	    if (!entry->header.deleted)
		new++;
    } else
	new = 0;
    dir_info->new_messages = new;
    if (dir_info->class == no_class &&
	!DAEMON(Client[i].uid) &&
	!NO_ACCESS(group_class(i, room_info.groups)))
	dir_info->class = room_info.default_class;
    if (dir_info->priority == DEFAULT_USER_OFFSET)
	dir_info->priority = room_info.room_offset;
    dir_info->priority += delta_priority(i, room_info.groups);
    if (dir_info->new_messages < 0)
	dir_info->new_messages = 0;
    if (Client[i].uid == room_info.owner_uid)
	dir_info->class = moderator;
    bcopy((char *)&room_info.groups,
	  (char *)&dir_info->groups,
	  sizeof(dir_info->groups));
    dir_info->archive = room_info.archive;
    return TRUE;
}

boolean update_cached_user_group_info(bb_uid uid, bb_gid gid)
{
    int slot;
    user_entry userinfo;
    group_entry group_info;

    if (!get_user_entry(&userinfo, uid)) {
	ERROR("error accessing user information");
	return FALSE;
    }
    if (!get_group_entry(&group_info, gid)) {
	ERROR("error accessing group information");
	return FALSE;
    }
    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on && Client[slot].uid == uid) {
	    bb_rid rid;

	    Client[slot].group_priority[gid] =
	      userinfo.group_priority[gid] + group_info.group_offset;
	    Client[slot].group_class[gid] = userinfo.group_class[gid];

	    for (rid = 0; rid <= Last_rid; rid++) {
		if (Client[slot].dir_info[rid] &&
		    GROUP_ISSET(gid, &Client[slot].dir_info[rid]->groups)) {
		    free(Client[slot].dir_info[rid]);
		    Client[slot].dir_info[rid] = NULL;
		    Client[slot].stamp[rid].update++;
		}
	    }
	    Client[slot].update_directory = 0;
	}
    return TRUE;
}

group_info *get_group_info(int i, bb_gid gid)
{
    group_entry groupentry;
    static group_info groupinfo;

    if (!get_group_entry(&groupentry, gid)) {
	ERROR("error accessing group information");
	return NULL;
    }
    groupinfo.gid = gid;
    bcopy(groupentry.name, groupinfo.name, sizeof(groupinfo.name));
    strcpy(groupinfo.owner, get_id(groupentry.owner));
    groupinfo.default_class = groupentry.default_class;
    groupinfo.max_expiration = groupentry.max_expiration;
    groupinfo.priority = Client[i].group_priority[gid];
    groupinfo.class = Client[i].group_class[gid];
    if (groupinfo.class == no_class)
	groupinfo.class = groupentry.default_class;
    if (Client[i].uid == groupentry.owner)
	groupinfo.class = moderator;
    return &groupinfo;
}

bb_priority delta_priority(int i, group_set groups)
{
    bb_gid gid;
    bb_priority priority = SYSTEM_CUTOFF;

    for (gid = 0; gid <= Last_gid; gid++)
	if (GROUP_ISSET(gid, &groups) &&
	    Client[i].group_priority[gid] > priority)
	    priority = Client[i].group_priority[gid];
    return priority;
}

bb_priority group_class(int i, group_set groups)
{
    bb_gid gid;
    class_t class = no_access, group_class;
    group_entry groupinfo;

    for (gid = 0; gid <= Last_gid; gid++) {
	if (GROUP_ISSET(gid, &groups)) {
	    group_class = Client[i].group_class[gid];
	    if (group_class == no_class && !DAEMON(Client[i].uid))
		group_class = get_group_entry(&groupinfo, gid) ?
		  groupinfo.default_class : no_access;
	    if (group_class > class)
		class = group_class;
	}
    }
    return class;
}

time_t expiration_limit(group_set groups)
{
    bb_gid gid;
    time_t max_expiration = 0;
    group_entry groupinfo;

    for (gid = 0; gid <= Last_gid; gid++) {
	if (GROUP_ISSET(gid, &groups))
	    if (get_group_entry(&groupinfo, gid) &&
		groupinfo.max_expiration > max_expiration)
		max_expiration = groupinfo.max_expiration;
    }
    return max_expiration;
}

int remove_room(bb_rid rid)
{
    directory_entry room_info;
    summary_entry *message;
    summary_list *summary;
    char filename[BUFFER_SIZE];
    bb_uid uid;

    if (Read_only) {
	ERROR("Messagebase is readonly.");
	return 0;
    }
    if (!get_directory_entry(&room_info, rid)) {
	ERROR("error accessing room information");
	return 0;
    }
    if (!(summary = get_summary(rid))) {
	ERROR("error accessing room summary");
	return 0;
    }
    for (message = summary->list;
	 message < summary->list + summary->count;
	 message++)
	expire_message(rid, message->number);
    sprintf(filename, ACTIVE_ROOM, rid);
    rmdir(filename);
    for (uid = 0; uid <= Last_uid; uid++)
	remove_room_user_entry(rid, uid);
    if (room_info.archive == common_archive) {
	FILE *room_log, *room_list;
	char *classes[] =
	{"no_access", "read_only", "read_write", "moderator", "no_class"};

	room_list = fopen(EXPIRED_ROOMS, "a");
	if (!room_list) {
	    syslog(LOG_ERR, "%s: %m", EXPIRED_ROOMS);
	} else {
	    fprintf(room_list, "%d %s\n", rid, room_info.name);
	    fclose(room_list);
	}
	sprintf(filename, EXPIRED_ROOM, rid);
	mkdir(filename, 0777);
	strcat(filename, "/Log");
	room_log = fopen(filename, "a");
	if (!room_log) {
	    syslog(LOG_ERR, "%s: %m", filename);
	} else {
	    fprintf(room_log, "Room ID:\t%d\nRoom Name:\t%s\nOwner:\t\t%d (%s)\nCreated:\t%sDefault Class:\t%s\nExpiration:\t%d (%d day%s)\nSummary:\t%s\nGroups:\t\t0x%04x\nPriority:\t%d\n",
		    rid,
		    room_info.name,
		    room_info.owner_uid, get_id(room_info.owner_uid),
		    ctime(&room_info.creation_time),
		    classes[room_info.default_class],
		    room_info.default_expiration,
		    room_info.default_expiration / 86400,
		    room_info.default_expiration / 86400 == 1 ? "" : "s",
		    room_info.summary_line,
		    *room_info.groups.group_bits,
		    room_info.room_offset);
	    fclose(room_log);
	}
    }
    if (!remove_rid(room_info.name)) {
	ERROR("error removing room identification");
	return 0;
    }
    if (!remove_directory_entry(rid)) {
	ERROR("error removing room information");
	return 0;
    }
    return 1;
}

int expire_message(bb_rid rid, int message)
{
    char filename[BUFFER_SIZE];

    if (Disable_expire) {
	ERROR("Expiration is disabled.");
	return 0;
    }
    if (Read_only) {
	ERROR("Messagebase is readonly.");
	return 0;
    }
    if (!archive(rid, message))
	return 0;
    remove_message_entry(rid, message);
    sprintf(filename, ACTIVE_MESSAGE, rid, message);
    if (unlink(filename) == -1) {
	syslog(LOG_ERR, "error deleting %s: %m\n", filename);
	return 0;
    }
    return 1;
}

static int archive(bb_rid rid, int message)
{
    directory_entry room_info;
    message_entry message_info;

    if (!db_get_message_entry(&message_info, rid, message))
	return 0;
    if(message_info.deleted) return 1;
    if (!get_directory_entry(&room_info, rid)) {
	ERROR("error accessing room information");
	return 0;
    }
    switch (room_info.archive) {
    case no_archive:
	break;
    case mail_archive:
	{
	    char subject[BUFFER_SIZE];

	    sprintf(subject, "archive of %s (%d): #%d",
		    room_info.name, rid, message);
	    if (!mail_message(rid, message, Owner_uid, room_info.owner_uid,
			      subject)) {
		syslog(LOG_ERR, "error archiving %d/%d to mail", rid, message);
		return 0;
	    }
	    break;
	}
    case common_archive:
    default:
	{
	    char filename[BUFFER_SIZE], newname[BUFFER_SIZE];
	    FILE *active_msg, *expired_msg;
	    int c;

	    sprintf(filename, ACTIVE_MESSAGE, rid, message);
	    sprintf(newname, EXPIRED_MESSAGE, rid, message);
	    expired_msg = fopen(newname, "w");
	    if (!expired_msg && errno == ENOENT) {
		char dirname[BUFFER_SIZE];

		sprintf(dirname, EXPIRED_ROOM, rid);
		mkdir(dirname, 0777);
		expired_msg = fopen(newname, "w");
	    }
	    if (!expired_msg) {
		syslog(LOG_ERR, "error writing %s: %m\n", newname);
		return 0;
	    }
	    fprintf(expired_msg,
		    "Date: %sFrom: %s<%s> (%d)\nSubject: %s\nLines: %d\n\n",
		    ctime(&message_info.date),
		    message_info.poster_nick,
		    message_info.poster_id,
		    message_info.poster_uid,
		    message_info.subject,
		    message_info.lines);
	    if (!(active_msg = fopen(filename, "r"))) {
		if (errno != ENOENT) {
		    syslog(LOG_ERR, "error reading %s: %m\n", filename);
		    return 0;
		}
	    } else {
		for (c = getc(active_msg); c != EOF; c = getc(active_msg))
		    putc((char)c, expired_msg);
		fclose(active_msg);
	    }
	    fclose(expired_msg);
	    break;
	}
    }
    return 1;
}

char *sh_lit(char *string)
{
    static char lit[BUFFER_SIZE];
    char *p, *q;

    q = lit;
    *q++ = '\'';
    for (p = string; *p; p++, q++) {
	switch (*p) {
	case '\'':
	    *q++ = '\'';
	    *q++ = '\\';
	    *q++ = *p;
	    *q = '\'';
	    break;
	case '\\':
	    *q++ = *p;
	default:
	    *q = *p;
	}
    }
    *q++ = '\'';
    *q = '\0';
    return lit;
}

FILE *mail(bb_uid from_uid, bb_uid to_uid, char *subject)
{
    char command[BUFFER_SIZE];
    FILE *message = NULL;
    char from_mail_address[BUFFER_SIZE], to_mail_address[BUFFER_SIZE];
    user_entry userinfo;

    if (!get_user_entry(&userinfo, from_uid) &&
	!get_user_entry(&userinfo, Owner_uid)) {
	ERROR("error accessing sender user information");
	return NULL;
    }
    strcpy(from_mail_address, *userinfo.mail_address ?
	   userinfo.mail_address : userinfo.last_access_source);
    if (!get_user_entry(&userinfo, to_uid)) {
	ERROR("error accessing recipient user information");
	return NULL;
    }
    strcpy(to_mail_address, *userinfo.mail_address ?
	   userinfo.mail_address : userinfo.last_access_source);
    sprintf(command, "/usr/lib/sendmail %s", sh_lit(to_mail_address));
    if (!(message = popen(command, "w"))) {
	syslog(LOG_ERR, "popen(\"%s\", \"w\")", command);
	ERROR(sys_errlist[errno]);
	return NULL;
    }
    fprintf(message, "From: <%s>\nTo: <%s>\nSubject: %s\n\n",
	    from_mail_address, to_mail_address, subject);
    return message;
}

void server_shutdown(boolean reboot, boolean immediate)
{
    extern int master_socket;
    int slot;

    syslog(LOG_INFO, "server %s %s",
	   reboot ? "rebooting" : "shutting down",
	   immediate ? "immediately" : "eventually");
    Server_Shutdown = TRUE;
    Server_Reboot = reboot;

    if (immediate) {
    /* Stop accepting connections */
	close(master_socket);
    /* Close all clients. */
	for (slot = 0; slot <= Max_active_slot; slot++) {
	    if (Client[slot].signed_on)
		save_room_pointer(slot);
	    if (Client[slot].inuse)
		close_client(slot);
	}
    }
}

signal_t terminate()
{
    signal(SIGTERM, terminate);
    server_shutdown(FALSE, TRUE);
    return (signal_t)0;
}

signal_t interrupt()
{
    extern int Current_client;

    signal(SIGINT, interrupt);
    if (Client[Current_client].inuse) {
	syslog(LOG_INFO, "kicking <%s> (%d)",
	       Client[Current_client].id,
	       Client[Current_client].uid);
	if (Client[Current_client].signed_on)
	    save_room_pointer(Current_client);
	close_client(Current_client);
    }
    return (signal_t)0;
}

signal_t hangup()
{
    signal(SIGHUP, hangup);
    server_shutdown(TRUE, TRUE);
    return (signal_t)0;
}

extern char **Argv;
char first_arg[] = "bb-server                                                                       ";

signal_t server_segv()
{
            *Argv = first_arg;
    syslog(LOG_INFO, "rebooting: segmentation violation - %m");
    execv(SERVER, Argv);
    syslog(LOG_INFO, "reboot failed - %s: %m", SERVER);
    return (signal_t)0;
}

signal_t server_reboot()
{
            *Argv = first_arg;
    syslog(LOG_INFO, "rebooting");
    execv(SERVER, Argv);
    syslog(LOG_INFO, "reboot failed - %s: %m", SERVER);
    return (signal_t)0;
}

void close_client(int slot)
{
    bb_rid rid;

 /* Close a client connection, marking its slot as available. */

    shutdown(Client[slot].socket, 2);
    close(Client[slot].socket);
    if (Client[slot].signed_on) {
	user_entry userinfo;

	if (get_user_entry(&userinfo, Client[slot].uid)) {
	    userinfo.connect_time += time(0) - Client[slot].signon_time;
	    userinfo.posts += Client[slot].posts;
	    put_user_entry(Client[slot].uid, &userinfo);
	}
    }
    Client[slot].inuse = FALSE;
    Client[slot].signed_on = FALSE;
    Client[slot].uid = -1;
    Client[slot].major_version = 0;
    Client[slot].minor_version = 0;
    Client[slot].socket = 0;
    Client[slot].errors = 0;
    if (Client[slot].dir_info) {
	for (rid = 0; rid <= Last_rid; rid++) {
	    if (Client[slot].dir_info[rid]) {
		free(Client[slot].dir_info[rid]);
		Client[slot].dir_info[rid] = NULL;
	    }
	}
	free(Client[slot].dir_info);
	Client[slot].dir_info = NULL;
    }
    if (slot < Free_slot)
	Free_slot = slot;
    if (slot == Max_active_slot) {
	for (; Max_active_slot >= 0; Max_active_slot--)
	    if (Client[Max_active_slot].inuse)
		break;
	if (Server_Shutdown)
	    syslog(LOG_DEBUG,
		   "Shutdown requested -- max active slot: %d -> %d",
		   slot, Max_active_slot);
    }
}

int remove_user(bb_uid uid)
{
    user_entry userinfo;
    bb_rid rid;

    if (!get_user_entry(&userinfo, uid)) {
	ERROR("error accessing user information");
	return 0;
    }
    for (rid = 0; rid <= Last_rid; rid++)
	remove_room_user_entry(rid, uid);
    if (!remove_user_entry(uid)) {
	syslog(LOG_ERR, "error removing user entry for %d <%s>",
	       uid, userinfo.id);
	ERROR("removing user_entry");
	return 0;
    }
    if (!remove_uid(userinfo.id)) {
	syslog(LOG_ERR, "error removing uid %d <%s>", uid, userinfo.id);
	ERROR("removing uid");
	return 0;
    }
    return 1;
}

int mail_message(bb_rid rid, int message,
		     bb_uid from_uid, bb_uid to_uid, char *subject)
{
    char filename[BUFFER_SIZE];
    message_entry message_info;
    FILE *active_msg, *mail_message;
    int c;

    if (!db_get_message_entry(&message_info, rid, message))
	return 0;
    sprintf(filename, ACTIVE_MESSAGE, rid, message);
    mail_message = mail(from_uid, to_uid, subject);
    if (!mail_message)
	return 0;
    fprintf(mail_message,
	    "Date: %sFrom: %s<%s> (%d)\nSubject: %s\nLines: %d\n\n",
	    ctime(&message_info.date),
	    message_info.poster_nick,
	    message_info.poster_id,
	    message_info.poster_uid,
	    message_info.subject,
	    message_info.lines);
    if (!(active_msg = fopen(filename, "r"))) {
	syslog(LOG_ERR, "error reading %s: %m\n", filename);
    } else {
	for (c = getc(active_msg); c != EOF; c = getc(active_msg))
	    putc((char)c, mail_message);
	fclose(active_msg);
    }
    pclose(mail_message);
    return active_msg != NULL;
}

static int summary_order(const void *p1, const void *p2)
{
    const summary_entry *se1 = p1, *se2 = p2;

    return se1->number - se2->number;
}

static void sort_summary(bb_rid room)
{
    summary_list *summary;

    if (!(summary = get_summary(room)))
	return;
    qsort(summary->list, summary->count, sizeof(*summary->list),
	  summary_order);
}

void expire_messages(bb_rid room)
{
    boolean room_info_changed = FALSE;
    directory_entry room_info;
    summary_list *tmp, summary;
    summary_entry *entry;
    int message, last_message = -1;

    if (Disable_expire)
	return;
    if (!get_directory_entry(&room_info, room))
	return;
    if (!(tmp = get_summary(room)))
	return;
    summary.count = tmp->count;
    if (!summary.count)
	return;
    summary.list = (summary_entry *)
      malloc(summary.count * sizeof(summary_entry));
    if (!summary.list)
	return;
    bcopy(tmp->list, summary.list, summary.count * sizeof(summary_entry));
    for (entry = summary.list;
	 entry < summary.list + summary.count;
	 entry++) {
	message = entry->number;
	if (message < last_message) {
	    syslog(LOG_WARNING,
		   "corrupted summary in %d (%s): #%d followed #%d",
		   room, get_roomname(room), message, last_message);
#ifdef FIXED
	    free(summary.list);
#endif FIXED
	    sort_summary(room);
#ifdef FIXED
	    expire_messages(room);
	    return;
#endif FIXED
	}
	if (message == last_message) {
	    syslog(LOG_WARNING, "corrupted summary in %d (%s): duplicate #%d",
		   room, get_roomname(room), last_message);
	    remove_message_entry(room, last_message);
	}
	last_message = message;
	if (entry->header.expiration < Now &&
	    expire_message(room, message)) {
	    if (message == room_info.first_message_number) {
		room_info.first_message_number++;
		room_info_changed = TRUE;
	    }
	}
    }
    free(summary.list);
    if (room_info_changed)
	put_directory_entry(room, &room_info);
}

void expire_room(bb_rid rid)
{
    directory_entry room_info;
    summary_list *summary;
    summary_entry *message;
    int slot;
    bb_gid gid;
    group_entry groupinfo;

    if (Disable_expire)
	return;
    if (!get_directory_entry(&room_info, rid))
	return;
    if (!(summary = get_summary(rid)))
	return;

    for (message = summary->list;
	 message < summary->list + summary->count;
	 message++)
	if (!message->header.deleted)
	    return;

 /*
  * Old_Room has no (undeleted) messages left.  Make a few saving throws; if
  * they fail, remove the room. 
  */

 /* someone is still in the room */
    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on && Client[slot].rid == rid)
	    return;

 /* the room is a member of a supported group */
    for (gid = 0; gid <= Last_gid; gid++)
	if (GROUP_ISSET(gid, &room_info.groups) &&
	    get_group_entry(&groupinfo, gid) &&
	    groupinfo.group_offset >= 0)
	    return;

 /* the room is restricted:  send the owner mail */
    if (room_info.default_class == no_access) {
	FILE *message_f;

	if (message_f = mail(Owner_uid, room_info.owner_uid, "empty room on bb")) {
	    fprintf(message_f, "The %s room is empty.  If you are no longer using this room,\nplease remove it.\n",
		    room_info.name);
	    pclose(message_f);
	}
	return;
    }
    remove_room(rid);
}
